/****************************************************************
 * Licensed to the Apache Software Foundation (ASF) under one   *
 * or more contributor license agreements.  See the NOTICE file *
 * distributed with this work for additional information        *
 * regarding copyright ownership.  The ASF licenses this file   *
 * to you under the Apache License, Version 2.0 (the            *
 * "License"); you may not use this file except in compliance   *
 * with the License.  You may obtain a copy of the License at   *
 *                                                              *
 *   http://www.apache.org/licenses/LICENSE-2.0                 *
 *                                                              *
 * Unless required by applicable law or agreed to in writing,   *
 * software distributed under the License is distributed on an  *
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY       *
 * KIND, either express or implied.  See the License for the    *
 * specific language governing permissions and limitations      *
 * under the License.                                           *
 ****************************************************************/

package org.apache.james.transport.matchers;

import java.io.File;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.annotation.Resource;
import javax.mail.MessagingException;
import javax.sql.DataSource;

import org.apache.james.services.FileSystem;
import org.apache.james.transport.mailets.WhiteListManager;
import org.apache.james.user.api.JamesUser;
import org.apache.james.user.api.UsersRepository;
import org.apache.james.util.sql.JDBCUtil;
import org.apache.james.util.sql.SqlResources;
import org.apache.mailet.Mail;
import org.apache.mailet.MailAddress;
import org.apache.mailet.base.GenericMatcher;

public abstract class AbstractSQLWhitelistMatcher extends GenericMatcher {


    /** The user repository for this mail server.  Contains all the users with inboxes
     * on this server.
     */
    private UsersRepository localusers;

    protected DataSource datasource;
    

    /**
     * Holds value of property sqlFile.
     */
    private File sqlFile;

     /**
     * Holds value of property sqlParameters.
     */
    private Map<String,String> sqlParameters = new HashMap<String,String>();


    
    @Resource(name="datasource")
    public void setDataSource(DataSource datasource) {
        this.datasource = datasource;
    }
    
    @Resource(name="localusersrepository")
    public void setUsersRepository(UsersRepository localusers) {
        this.localusers = localusers;
    }
    
    /**
     * Getter for property sqlParameters.
     * @return Value of property sqlParameters.
     */
    private Map<String,String> getSqlParameters() {
        return this.sqlParameters;
    }
    
    /**
     * The JDBCUtil helper class
     */
    protected final JDBCUtil theJDBCUtil = new JDBCUtil() {
        protected void delegatedLog(String logString) {
            log(getMatcherName() +": " + logString);
        }
    };
    
    /**
     * Contains all of the sql strings for this component.
     */
    protected SqlResources sqlQueries = new SqlResources();

    private FileSystem fs;

    @Resource(name="filesystem")
    public void setFilesystem(FileSystem fs) {
        this.fs = fs;
    }
    
    @Override
    public void init() throws MessagingException {
        String repositoryPath = null;
        StringTokenizer st = new StringTokenizer(getCondition(), ", \t", false);
        if (st.hasMoreTokens()) {
            repositoryPath = st.nextToken().trim();
        }
        if (repositoryPath != null) {
            log("repositoryPath: " + repositoryPath);
        }
        else {
            throw new MessagingException("repositoryPath is null");
        }

      

        try {
            initSqlQueries(datasource.getConnection(), getMailetContext());
        } catch (Exception e) {
            throw new MessagingException("Exception initializing queries", e);
        }        
        
        super.init();
	}

    public Collection<MailAddress> match(Mail mail) throws MessagingException {
        // check if it's a local sender
        MailAddress senderMailAddress = mail.getSender();
        if (senderMailAddress == null) {
            return null;
        }
        if (getMailetContext().isLocalEmail(senderMailAddress)) {
            // is a local sender, so return
            return null;
        }

        String senderUser = senderMailAddress.getLocalPart();
        String senderHost = senderMailAddress.getDomain();

        senderUser = senderUser.toLowerCase(Locale.US);
        senderHost = senderHost.toLowerCase(Locale.US);

        Collection<MailAddress> recipients = mail.getRecipients();

        Collection<MailAddress> inWhiteList = new java.util.HashSet<MailAddress>();

        for (Iterator<MailAddress> i = recipients.iterator(); i.hasNext();) {
            MailAddress recipientMailAddress = i.next();
            String recipientUser = recipientMailAddress.getLocalPart().toLowerCase(Locale.US);
            String recipientHost = recipientMailAddress.getDomain().toLowerCase(Locale.US);

            if (!getMailetContext().isLocalServer(recipientHost)) {
                // not a local recipient, so skip
                continue;
            }

            recipientUser = getPrimaryName(recipientUser);

            if (matchedWhitelist(recipientMailAddress, mail)) {
                // This address was already in the list
                inWhiteList.add(recipientMailAddress);
            }

        }

        return inWhiteList;

    }
    
    protected abstract boolean matchedWhitelist(MailAddress recipient, Mail mail) throws MessagingException;
    
	/** Gets the main name of a local customer, handling alias */
    protected String getPrimaryName(String originalUsername) {
        String username;
        try {
            username = localusers.getRealName(originalUsername);
            JamesUser user = (JamesUser) localusers.getUserByName(username);
            if (user.getAliasing()) {
                username = user.getAlias();
            }
        }
        catch (Exception e) {
            username = originalUsername;
        }
        return username;
    }
    
    /**
     * Initializes the sql query environment from the SqlResources file.
     * Will look for conf/sqlResources.xml.
     * Will <I>not</I> create the database resources, if missing
     * (this task is done, if needed, in the {@link WhiteListManager}
     * initialization routine).
     * @param conn The connection for accessing the database
     * @param mailetContext The current mailet context,
     * for finding the conf/sqlResources.xml file
     * @throws Exception If any error occurs
     */
    protected void initSqlQueries(Connection conn, org.apache.mailet.MailetContext mailetContext) throws Exception {
        try {
            if (conn.getAutoCommit()) {
                conn.setAutoCommit(false);
            }
            
            this.sqlFile = fs.getFile("classpath:sqlResources.xml");
            sqlQueries.init(this.sqlFile, getSQLSectionName(), conn, getSqlParameters());
            checkTables(conn);
        } finally {
            theJDBCUtil.closeJDBCConnection(conn);
        }
    }
    

    protected abstract String getTableName();
    
    protected abstract String getTableCreateQueryName();
    
    private void checkTables(Connection conn) throws SQLException {

        // Need to ask in the case that identifiers are stored, ask the DatabaseMetaInfo.
        // Try UPPER, lower, and MixedCase, to see if the table is there.
        
        boolean dbUpdated = false;
        
        dbUpdated = createTable(conn, getTableName(), getTableCreateQueryName());
        
        //Commit our changes if necessary.
        if (conn != null && dbUpdated && !conn.getAutoCommit()) {
            conn.commit();
            dbUpdated = false;
        }
            
    }
    
    private boolean createTable(Connection conn, String tableNameSqlStringName, String createSqlStringName) throws SQLException {
        String tableName = sqlQueries.getSqlString(tableNameSqlStringName, true);
        
        DatabaseMetaData dbMetaData = conn.getMetaData();

        // Try UPPER, lower, and MixedCase, to see if the table is there.
        if (theJDBCUtil.tableExists(dbMetaData, tableName)) {
            return false;
        }
        
        PreparedStatement createStatement = null;
        
        try {
            createStatement =
                    conn.prepareStatement(sqlQueries.getSqlString(createSqlStringName, true));
            createStatement.execute();
            
            StringBuffer logBuffer = null;
            logBuffer =
                    new StringBuffer(64)
                    .append("Created table '")
                    .append(tableName)
                    .append("' using sqlResources string '")
                    .append(createSqlStringName)
                    .append("'.");
            log(logBuffer.toString());
            
        } finally {
            theJDBCUtil.closeJDBCStatement(createStatement);
        }
        
        return true;
    }
    
    
    protected abstract String getSQLSectionName();
}
